==> cds4j-codegen-1.32.1-sources.jar: com/sap/cds/generator/util/TypeUtils.java <==
package com.sap.cds.generator.util;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.lang.model.element.Modifier;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sap.cds.Struct;
import com.sap.cds.generator.Configuration;
import com.sap.cds.reflect.CdsArrayedType;
import com.sap.cds.reflect.CdsAssociationType;
import com.sap.cds.reflect.CdsBaseType;
import com.sap.cds.reflect.CdsDefinition;
import com.sap.cds.reflect.CdsElement;
import com.sap.cds.reflect.CdsEntity;
import com.sap.cds.reflect.CdsModel;
import com.sap.cds.reflect.CdsSimpleType;
import com.sap.cds.reflect.CdsStructuredType;
import com.sap.cds.reflect.CdsType;
import com.sap.cds.util.CdsModelUtils;
import com.sap.cds.util.OnConditionAnalyzer;
import com.squareup.javapoet.ClassName;
import com.squareup.javapoet.CodeBlock;
import com.squareup.javapoet.MethodSpec;
import com.squareup.javapoet.ParameterSpec;
import com.squareup.javapoet.ParameterizedTypeName;
import com.squareup.javapoet.TypeName;
import com.squareup.javapoet.TypeSpec;

import static com.sap.cds.generator.util.NamesUtils.namespace;
import static com.sap.cds.generator.util.NamesUtils.qualifiedJavaClass;
import static com.sap.cds.generator.util.NamesUtils.unqualifiedName;
import static com.sap.cds.generator.util.PoetTypeName.getArrayTypeName;
import static com.sap.cds.generator.util.PoetTypeName.getTypeFromCdsName;
import static com.sap.cds.generator.util.PoetTypeName.getTypeName;
import static com.sap.cds.generator.writer.CaseFormatHelper.toUpperCamel;

public class TypeUtils {

	private static final Logger logger = LoggerFactory.getLogger(TypeUtils.class);

	private TypeUtils() {
	}

	public static TypeName getAttributeType(CdsType type, Configuration cfg) {
		return getAttributeType(type, cfg, null);
	}

	private static TypeName getAttributeType(CdsType type, Configuration cfg, String structTypeName) {
		TypeName attributeType;

		if (type.isAssociation()) {
			CdsAssociationType assocType = type.as(CdsAssociationType.class);
			CdsType target = assocType.getTargetAspect().orElseGet(assocType::getTarget);
			attributeType = getTypeFromCdsName(cfg.getBasePackage(), target.getQualifiedName(), target.getName());
			if (!CdsModelUtils.isSingleValued(type)) {
				attributeType = listOf(attributeType);
			}
		} else if (type.isSimple()) {
			attributeType = getTypeName(type.as(CdsSimpleType.class).getJavaType().getTypeName());
		} else if (type.isStructured()) {
			String targetName = type.getQualifiedName();

			// arrayed inlined structured type
			if (targetName.isEmpty() && type.getName().isEmpty() && structTypeName != null) {
				attributeType = getTypeName(toUpperCamel(structTypeName));
			} else {
				attributeType = getTypeFromCdsName(cfg.getBasePackage(), targetName, type.getName());
			}
		} else if (type.isArrayed()) {
			CdsType itemsType = type.as(CdsArrayedType.class).getItemsType();
			//Arrayed Types which are part of inline defined structured type will not have multiple segment qualifiedName.
			//For such types, we run the else condition. 
			boolean arrayedInlinedStructuredType = type.getQualifiedName().contains(".");
			TypeName innerAttributeType = getAttributeType(itemsType, cfg, !arrayedInlinedStructuredType ? type.getName() : null);
			if (innerAttributeType.toString().endsWith(".")) {
				// contains anonymous structured type with empty qualified name
				String className = toUpperCamel(unqualifiedName(type.getQualifiedName()));
				TypeName outer = getTypeName(
						qualifiedJavaClass(cfg.getBasePackage(), namespace(type.getQualifiedName())));
				attributeType = getArrayTypeName(getTypeName(outer + "." + className));
			} else {
				attributeType = getArrayTypeName(innerAttributeType);
			}
		} else {
			logger.warn("Interface Generation: Unsupported CDS Element with attribute name '" + type.getName()
					+ "' and type '" + type + "'");
			return null;
		}
		return attributeType;
	}

	public static TypeName getReturnType(CdsElement attribute, Configuration cfg) {
		boolean isMedia = attribute.annotations().anyMatch(a -> "Core.MediaType".equals(a.getName()));
		if (isMedia && attribute.getType().isSimple()) {
			CdsSimpleType simpleType = attribute.getType().as(CdsSimpleType.class);
			return PoetTypeName.getTypeName(CdsBaseType.cdsJavaMediaType(simpleType.getType()).getName());
		}
		if (attribute.getType().isStructured() && attribute.getType().getQualifiedName().isEmpty()) {
			return getTypeName(toUpperCamel(attribute.getName()));
		} else if (isAnonymousAspect(attribute)) {
			TypeName attributeType;
			attributeType = getTypeName(toUpperCamel(attribute.getName()));
			if (!CdsModelUtils.isSingleValued(attribute.getType())) {
				attributeType = listOf(attributeType);
			}
			return attributeType;
		}
		return getAttributeType(attribute.getType(), cfg);
	}

	public static ParameterizedTypeName listOf(TypeName type) {
		return ParameterizedTypeName.get(ClassName.get(List.class), type);
	}

	public static ClassName className(String name) {
		return ClassName.bestGuess(toUpperCamel(name));
	}

	public static ClassName builderClassName(CdsDefinition def) {
		return ClassName.bestGuess(toUpperCamel(def.getName()) + "_");
	}

	public static boolean isAnonymousAspect(CdsElement element) {
		CdsType type = element.getType();
		if (type.isAssociation()) {
			Optional<CdsType> targetAspect = type.as(CdsAssociationType.class).getTargetAspect();
			if (targetAspect.isPresent() && targetAspect.get().getQualifiedName().isEmpty()) {
				return true;
			}
		}
		return false;
	}

	public static boolean isCompositionOfAspectsWithStructuredElements(CdsElement element) {
		return element.getType().as(CdsAssociationType.class).getTargetAspect()
				.map(ta -> ta.as(CdsStructuredType.class).elements().anyMatch(e -> e.getType().isStructured()))
				.orElse(false);
	}

	public static Stream<CdsElement> getAnonymousElements(CdsElement element) {
		CdsType type = element.getType();
		if (type.isAssociation()) {
			Optional<CdsType> targetAspect = type.as(CdsAssociationType.class).getTargetAspect();
			if (targetAspect.isPresent() && targetAspect.get().getQualifiedName().isEmpty()) {
				return targetAspect.get().as(CdsStructuredType.class).elements();
			}
		}
		return Stream.empty();
	}

	public static void addStaticMethod(TypeSpec.Builder builder, TypeName returnType,
			TypeSpec.Builder... anonymousbuilder) {
		MethodSpec.Builder methodBuilder = MethodSpec.methodBuilder("create").returns(returnType)
				.addModifiers(Modifier.PUBLIC, Modifier.STATIC);
		CodeBlock.Builder codeBuilder = CodeBlock.builder();
		codeBuilder.addStatement("return $T.create($T.class)", ClassName.get(Struct.class), returnType);

		methodBuilder.addCode(codeBuilder.build());
		if (anonymousbuilder.length > 0) {
			anonymousbuilder[0].addMethod(methodBuilder.build());
		} else {
			builder.addMethod(methodBuilder.build());
		}
	}

	public static void addStaticCreateForKeys(TypeSpec.Builder builder, CdsEntity entity, Map<String, ParameterSpec> parameters) {
		if (!parameters.isEmpty()) {
			ClassName returnType = className(entity.getName());
			MethodSpec.Builder methodBuilder = MethodSpec.methodBuilder("create").returns(returnType)
														 .addModifiers(Modifier.PUBLIC, Modifier.STATIC);
			parameters.forEach((name, spec) -> methodBuilder.addParameter(spec));

			CodeBlock.Builder codeBuilder = CodeBlock.builder();
			codeBuilder.addStatement("$T<String, Object> keys = new $T<>()",
									 ClassName.get(Map.class),
									 ClassName.get(HashMap.class));
			parameters.forEach((name, spec) ->
					codeBuilder.addStatement("keys.put($L, $L)", name, spec.name));
			codeBuilder.addStatement("return $T.access(keys).as($T.class)", ClassName.get(Struct.class), returnType);

			methodBuilder.addCode(codeBuilder.build());
			builder.addMethod(methodBuilder.build());
		}
	}


	public static List<CdsElement> getManagedToOneFks(CdsElement attribute) {
		return new OnConditionAnalyzer(attribute, false).getFkMapping().keySet().stream()
				.map(m -> attribute.getDeclaringType().as(CdsStructuredType.class).findElement(m)).map(Optional::get)
				.collect(Collectors.toList());
	}

	public static void logWarningForManyToManyWithStructElement(CdsModel model, CdsStructuredType struct) {
		if (struct.elements().anyMatch(e -> e.getType().isStructured())) {
			Optional<CdsElement> assocElement = model.entities().flatMap(CdsStructuredType::associations)
					.filter(a -> a.getType().as(CdsAssociationType.class).getTargetAspect()
							.map(ta -> ta.getName().equals(struct.getName())).orElse(false))
					.findFirst();
			if (assocElement.isPresent()) {
				logger.warn(
						"Limited CRUD operation support available with composition of aspects(" + struct.getQualifiedName() + ") containing structured elements. Use dynamic queries in such cases.");
			}
		}
	}
}

==> cds4j-codegen-2.4.0-sources.jar: com/sap/cds/generator/util/TypeUtils.java <==
package com.sap.cds.generator.util;

import static com.sap.cds.generator.util.NamesUtils.namespace;
import static com.sap.cds.generator.util.NamesUtils.qualifiedJavaClass;
import static com.sap.cds.generator.util.NamesUtils.unqualifiedName;
import static com.sap.cds.generator.util.PoetTypeName.getArrayTypeName;
import static com.sap.cds.generator.util.PoetTypeName.getTypeFromCdsName;
import static com.sap.cds.generator.util.PoetTypeName.getTypeName;
import static com.sap.cds.generator.writer.CaseFormatHelper.toUpperCamel;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.lang.model.element.Modifier;

import com.sap.cds.Struct;
import com.sap.cds.generator.Configuration;
import com.sap.cds.reflect.CdsArrayedType;
import com.sap.cds.reflect.CdsAssociationType;
import com.sap.cds.reflect.CdsBaseType;
import com.sap.cds.reflect.CdsDefinition;
import com.sap.cds.reflect.CdsElement;
import com.sap.cds.reflect.CdsEntity;
import com.sap.cds.reflect.CdsModel;
import com.sap.cds.reflect.CdsSimpleType;
import com.sap.cds.reflect.CdsStructuredType;
import com.sap.cds.reflect.CdsType;
import com.sap.cds.util.CdsModelUtils;
import com.sap.cds.util.OnConditionAnalyzer;
import com.squareup.javapoet.ClassName;
import com.squareup.javapoet.CodeBlock;
import com.squareup.javapoet.MethodSpec;
import com.squareup.javapoet.ParameterSpec;
import com.squareup.javapoet.ParameterizedTypeName;
import com.squareup.javapoet.TypeName;
import com.squareup.javapoet.TypeSpec;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class TypeUtils {

	private static final Logger logger = LoggerFactory.getLogger(TypeUtils.class);

	private TypeUtils() {
	}

	public static TypeName getAttributeType(CdsType type, Configuration cfg) {
		return getAttributeType(type, cfg, null);
	}

	private static TypeName getAttributeType(CdsType type, Configuration cfg, String structTypeName) {
		TypeName attributeType;

		if (type.isAssociation()) {
			CdsAssociationType assocType = type.as(CdsAssociationType.class);
			CdsType target = assocType.getTargetAspect().orElseGet(assocType::getTarget);
			attributeType = getTypeFromCdsName(cfg.getBasePackage(), target.getQualifiedName(), target.getName());
			if (!CdsModelUtils.isSingleValued(type)) {
				attributeType = listOf(attributeType);
			}
		} else if (type.isSimple()) {
			attributeType = getTypeName(type.as(CdsSimpleType.class).getJavaType().getTypeName());
		} else if (type.isStructured()) {
			String targetName = type.getQualifiedName();

			// arrayed inlined structured type
			if (targetName.isEmpty() && type.getName().isEmpty() && structTypeName != null) {
				attributeType = getTypeName(toUpperCamel(structTypeName));
			} else {
				attributeType = getTypeFromCdsName(cfg.getBasePackage(), targetName, type.getName());
			}
		} else if (type.isArrayed()) {
			CdsType itemsType = type.as(CdsArrayedType.class).getItemsType();
			//Arrayed Types which are part of inline defined structured type will not have multiple segment qualifiedName.
			//For such types, we run the else condition. 
			boolean arrayedInlinedStructuredType = type.getQualifiedName().contains(".");
			TypeName innerAttributeType = getAttributeType(itemsType, cfg, !arrayedInlinedStructuredType ? type.getName() : null);
			if (innerAttributeType.toString().endsWith(".")) {
				// contains anonymous structured type with empty qualified name
				String className = toUpperCamel(unqualifiedName(type.getQualifiedName()));
				TypeName outer = getTypeName(
						qualifiedJavaClass(cfg.getBasePackage(), namespace(type.getQualifiedName())));
				attributeType = getArrayTypeName(getTypeName(outer + "." + className));
			} else {
				attributeType = getArrayTypeName(innerAttributeType);
			}
		} else {
			logger.warn("Interface Generation: Unsupported CDS Element with attribute name '" + type.getName()
					+ "' and type '" + type + "'");
			return null;
		}
		return attributeType;
	}

	public static TypeName getReturnType(CdsElement attribute, Configuration cfg) {
		boolean isMedia = attribute.annotations().anyMatch(a -> "Core.MediaType".equals(a.getName()));
		if (isMedia && attribute.getType().isSimple()) {
			CdsSimpleType simpleType = attribute.getType().as(CdsSimpleType.class);
			return PoetTypeName.getTypeName(CdsBaseType.cdsJavaMediaType(simpleType.getType()).getName());
		}
		if (attribute.getType().isStructured() && attribute.getType().getQualifiedName().isEmpty()) {
			return getTypeName(toUpperCamel(attribute.getName()));
		} else if (isAnonymousAspect(attribute)) {
			TypeName attributeType;
			attributeType = getTypeName(toUpperCamel(attribute.getName()));
			if (!CdsModelUtils.isSingleValued(attribute.getType())) {
				attributeType = listOf(attributeType);
			}
			return attributeType;
		}
		return getAttributeType(attribute.getType(), cfg);
	}

	public static TypeName getOperationResultType(CdsDefinition operation, CdsType returnType, NamesUtils namesUtils) {
		if (TypeUtils.isAnonymousType(returnType)) {
			final String RETURN_TYPE_SUFFIX = "ReturnType";
			return getInnerEventContextClassName(operation, RETURN_TYPE_SUFFIX);
		}

		if (returnType.isStructured()) {
			return getTypeName(namesUtils.qualifiedJavaClassName(returnType));
		} else if (returnType.isSimple()) {
			return getTypeName(returnType.as(CdsSimpleType.class).getJavaType().getTypeName());
		} else if (returnType.isArrayed()) {
			TypeName nestedType = getOperationResultType(operation, returnType.as(CdsArrayedType.class).getItemsType(), namesUtils);
			return ParameterizedTypeName.get(ClassName.get(Collection.class), nestedType);
		} else {
			logger.warn("Consumption Interface Generation: Unsupported CDS Element with attribute name {} and type {}",
					operation.getName(), returnType.getName());
			return null;
		}
	}

	public static ClassName getInnerEventContextClassName(CdsDefinition def, String innerClassName) {
		return ClassName.bestGuess(String.format("%s.%s", toUpperCamel(def.getName() + "Context"), toUpperCamel(innerClassName)));
	}

	public static ParameterizedTypeName listOf(TypeName type) {
		return ParameterizedTypeName.get(ClassName.get(List.class), type);
	}

	public static ClassName className(String name) {
		return ClassName.bestGuess(toUpperCamel(name));
	}

	public static ClassName builderClassName(CdsDefinition def) {
		return ClassName.bestGuess(toUpperCamel(def.getName()) + "_");
	}

	public static boolean isAnonymousAspect(CdsElement element) {
		CdsType type = element.getType();
		if (type.isAssociation()) {
			Optional<CdsType> targetAspect = type.as(CdsAssociationType.class).getTargetAspect();
			if (targetAspect.isPresent() && targetAspect.get().getQualifiedName().isEmpty()) {
				return true;
			}
		}
		return false;
	}

	public static boolean isAnonymousType(CdsType type) {
		if (type == null) {
			return false;
		} else if (type.isArrayed()) {
			return isAnonymousType(type.as(CdsArrayedType.class).getItemsType());
		} else if (type.isStructured() && type.getQualifiedName().isEmpty()) {
			return true;
		}
		return false;
	}

	public static boolean isCompositionOfAspectsWithStructuredElements(CdsElement element) {
		return element.getType().as(CdsAssociationType.class).getTargetAspect()
				.map(ta -> ta.as(CdsStructuredType.class).elements().anyMatch(e -> e.getType().isStructured()))
				.orElse(false);
	}

	public static Stream<CdsElement> getAnonymousElements(CdsElement element) {
		CdsType type = element.getType();
		if (type.isAssociation()) {
			Optional<CdsType> targetAspect = type.as(CdsAssociationType.class).getTargetAspect();
			if (targetAspect.isPresent() && targetAspect.get().getQualifiedName().isEmpty()) {
				return targetAspect.get().as(CdsStructuredType.class).elements();
			}
		}
		return Stream.empty();
	}

	public static void addStaticMethod(TypeSpec.Builder builder, TypeName returnType,
			TypeSpec.Builder... anonymousbuilder) {
		MethodSpec.Builder methodBuilder = MethodSpec.methodBuilder("create").returns(returnType)
				.addModifiers(Modifier.PUBLIC, Modifier.STATIC);
		CodeBlock.Builder codeBuilder = CodeBlock.builder();
		codeBuilder.addStatement("return $T.create($T.class)", ClassName.get(Struct.class), returnType);

		methodBuilder.addCode(codeBuilder.build());
		if (anonymousbuilder.length > 0) {
			anonymousbuilder[0].addMethod(methodBuilder.build());
		} else {
			builder.addMethod(methodBuilder.build());
		}
	}

	public static void addStaticCreateForKeys(TypeSpec.Builder builder, CdsEntity entity, Map<String, ParameterSpec> parameters) {
		if (!parameters.isEmpty()) {
			ClassName returnType = className(entity.getName());
			MethodSpec.Builder methodBuilder = MethodSpec.methodBuilder("create").returns(returnType)
														 .addModifiers(Modifier.PUBLIC, Modifier.STATIC);
			parameters.forEach((name, spec) -> methodBuilder.addParameter(spec));

			CodeBlock.Builder codeBuilder = CodeBlock.builder();
			codeBuilder.addStatement("$T<String, Object> keys = new $T<>()",
									 ClassName.get(Map.class),
									 ClassName.get(HashMap.class));
			parameters.forEach((name, spec) ->
					codeBuilder.addStatement("keys.put($L, $L)", name, spec.name));
			codeBuilder.addStatement("return $T.access(keys).as($T.class)", ClassName.get(Struct.class), returnType);

			methodBuilder.addCode(codeBuilder.build());
			builder.addMethod(methodBuilder.build());
		}
	}


	public static List<CdsElement> getManagedToOneFks(CdsElement attribute) {
		return new OnConditionAnalyzer(attribute, false).getFkMapping().keySet().stream()
				.map(m -> attribute.getDeclaringType().as(CdsStructuredType.class).findElement(m)).map(Optional::get)
				.collect(Collectors.toList());
	}

	public static void logWarningForManyToManyWithStructElement(CdsModel model, CdsStructuredType struct) {
		if (struct.elements().anyMatch(e -> e.getType().isStructured())) {
			Optional<CdsElement> assocElement = model.entities().flatMap(CdsStructuredType::associations)
					.filter(a -> a.getType().as(CdsAssociationType.class).getTargetAspect()
							.map(ta -> ta.getName().equals(struct.getName())).orElse(false))
					.findFirst();
			if (assocElement.isPresent()) {
				logger.warn(
						"Limited CRUD operation support available with composition of aspects(" + struct.getQualifiedName() + ") containing structured elements. Use dynamic queries in such cases.");
			}
		}
	}
}
